// Copyright (C) 2014, Ives Rey-Otero, CMLA ENS Cachan
// // <ives.rey-otero@cmla.ens-cachan.fr>
// //
// // Version 20150806 (August 6th, 2015)
// //
// //
// // This program is free software: you can use, modify and/or
// // redistribute it under the terms of the GNU General Public
// // License as published by the Free Software Foundation, either
// // version 3 of the License, or (at your option) any later
// // version. You should have received a copy of this license along
// // this program. If not, see <http://www.gnu.org/licenses/>.
//
//
//



#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <fftw3.h>

#define MAX(i,j) ( (i)<(j) ? (j):(i) )
#define MIN(i,j) ( (i)<(j) ? (i):(j) )

#ifndef M_PI
    #define M_PI 3.14159265359
#endif


#include "lib_gaussconv.h"

// Allocate memory or abort on failure
void* xmalloc(size_t size)
{
    if (size == 0)
        fprintf(stderr,"xmalloc: zero size");
    void *p = malloc(size);
    if (!p)
    {
        float sm = size / (0x100000 * 1.0);
        fprintf(stderr,"xmalloc: out of memory when requesting "
                "%zu bytes (%gMB)",//:\"%s\"",
                size, sm);//, strerror(errno));
    }
    return p;
}


// Free memory allocated by xmalloc
void xfree(void* p)
{
    if (!p)
        fprintf(stderr,"trying to free a null pointer");
    free(p);
}

////////  Sampled Gaussian kernel - discrete convolution //////////




/** @brief Build mono-dimensional sampled Gaussian kernel
 *   of size 2*rad+1 and standard deviation sigma
 */
static float* malloc_gaussian_kernel(float sigma, int radius)
{
    float* kernel = xmalloc((2*radius+1)*sizeof(float));
    float sum = 0;
    if(sigma>0){
        for(int i = -radius; i <= radius; i++){
            kernel[radius+i] = exp(-0.5*(float)i*(float)i/sigma/sigma);
            sum += kernel[radius+i];
        }
        for(int i = -radius; i <= radius; i++)
            kernel[radius+i] /= sum;
    }else{
       for(int i = -radius; i <= radius; i++){
           kernel[radius+i] = 0.0;
       }
       kernel[radius] = 1.;
    }
    return kernel;
}




/** @brief Apply a convolution with a separable kernel
 * and signal extension by symmetrization
 *
 * @param in             Input image of w X h samples
 * @param out            Output image (same dimension)
 *
 * @param xker           Kernel applied along direction x
 *                          L = r_xker
 *                          w = 2*r_xker+1
 *
 * @param yker           Kernel applied along direction y
 *                          L = r_yker
 *                          w = 2*r_yker+1
 *
 *  Compute:
 *
 *   out[i,j] = \sum_{-r_xker \leq k \leq +r_xker}
 *                    xker[k]
 *                    . \sum_{-r_xker \leq l \leq +r_xker}
 *                              yker[l].in[i-k,j-l]
 *
 *  Border condition:  symmetrization at border
 *                     (at x = -1./2 and x = w-1./2)
 *
 */
static void convolve(float* in, float* out, int w, int h,
              float* xker, int r_xker,
              float* yker, int r_yker){
    int i,j;
    int i_p,j_p;
    int k;
    float tmp;
    float* im_tmp = xmalloc(w*h*sizeof(float));
    /* convolution along x coordinates */
    for(i=0;i<h;i++){
        for(j=0;j<w;j++){
            tmp=0;
            for(k=-r_xker;k<=r_xker;k++){
                i_p = (i-k+2*h)%(2*h);
                if(i_p>h-1){i_p=2*h-1-i_p;}
                tmp += xker[r_xker+k]*in[i_p*w+j];
            }
            im_tmp[i*w+j]=tmp;
        }
    }
    /* convolution along y coordinates */
    for(i=0;i<h;i++){
        for(j=0;j<w;j++){
            tmp=0;
            for(k=-r_yker;k<=r_yker;k++){
                j_p = (j-k+2*w)%(2*w);
                if(j_p>w-1){j_p=2*w-1-j_p;}
                tmp += yker[r_yker+k]*im_tmp[i*w+j_p];
            }
            out[i*w+j]=tmp;
        }
    }
    xfree(im_tmp);
}


/**
 *  Apply discrete convolution to the image with a truncated sampled Gaussian kernel
 *
 *
 */
void gaussconv_sampled_kernel(float* in, float* out, int w, int h, float sigma, int k)
{
    assert(sigma>=0);
    int r_gker = (int)(k*sigma);
    float* gker = malloc_gaussian_kernel(sigma, r_gker);
    convolve(in,out,w,h,gker,r_gker,gker,r_gker);
    xfree(gker);
}



////////  DFT convolution - Gaussian convolution with the DFT interpolation  //






/**
 *  Apply the direct DFT transform of a 2D real valued image.
 *
 *  dft : size (w/2+1)*h
 *
 */
static void direct_dft(float* im, fftwf_complex* dft, int w, int h)
{
    fftwf_plan r2c;
	r2c = fftwf_plan_dft_r2c_2d(h, w, im, dft, FFTW_ESTIMATE);
	fftwf_execute(r2c);
    float norm = (float)(w*h);
    for(int i=0; i<(w/2+1)*h; i++){
        dft[i][0] /=norm;
        dft[i][1] /=norm;
    }
    fftwf_destroy_plan(r2c);
    fftwf_cleanup();
}


/**
 *  Apply the inverse DFT transform.
 *
 *  dft : size (w/2+1)*h
 *
 */
static void inverse_dft(fftwf_complex* dft, float* im, int w, int h)
{
    fftwf_plan c2r;
	c2r = fftwf_plan_dft_c2r_2d(h, w, dft, im, FFTW_ESTIMATE);
	fftwf_execute(c2r);
    fftwf_destroy_plan(c2r);
    fftwf_cleanup();
}


/**
 *
 *  Weighting the DFT coefficients of an image
 *   using the Fourier transform of the Gaussian function.
 *
 *  dft : size (w/2+1)*h
 *
 */
static void gaussian_on_dft(fftwf_complex* dftin, fftwf_complex* dftout, int w, int h, float sigma)
{
    float s = 2*sigma*sigma*M_PI*M_PI;
	for(int m = 0; m < h; m++){ //
        int mp = m;
		if (m > -h/2+h-1){mp = m-h;}
        float tmp_m = (float)mp/(float)h;  // frequency in the vertical direction
		for(int n = 0; n < w/2+1; n++){
			int np = n;
			if (n > -w/2+w-1){ np = n-w;}
            float tmp_n = (float)np/(float)w;  // frequency in the horizontal direction
			float weight = exp(-s*(tmp_m*tmp_m + tmp_n*tmp_n));
			dftout[m*(w/2+1)+n][0] = dftin[m*(w/2+1)+n][0] * weight;
			dftout[m*(w/2+1)+n][1] = dftin[m*(w/2+1)+n][1] * weight;
		}
	}
}

/**
 *
 * Continuous Gaussian convolution of the DFT interpolate.
 *
 */
void gaussconv_dft(float* in, float* out, int w, int h, float sigma)
{
	size_t fft_size = (w/2+1)*h;   /// fftw takes advantages of DFT conjugate symmetrie
	fftwf_complex* dftin  = (fftwf_complex *) fftwf_malloc(fft_size * sizeof(fftwf_complex));
	fftwf_complex* dftout = (fftwf_complex *) fftwf_malloc(fft_size * sizeof(fftwf_complex));

    direct_dft(in, dftin, w, h);
    gaussian_on_dft(dftin, dftout, w, h, sigma);
    inverse_dft(dftout, out, w, h);

    fftwf_free(dftin);
    fftwf_free(dftout);
}







/**
 *
 * Computation of the DCT coefficients
 *
 */
static void direct_dct(float* im, float* dct, int w, int h)
{
    fftwf_plan r2r;
    r2r = fftwf_plan_r2r_2d(h, w, im, dct, FFTW_REDFT10, FFTW_REDFT10, FFTW_ESTIMATE);
    fftwf_execute(r2r);
    float norm = (float)(2*w*2*h);
    for(int i=0; i<w*h; i++)
        dct[i] /=norm;
    fftwf_destroy_plan(r2r);
    fftwf_cleanup();
}


/**
 *
 * Computation of the IDCT coefficients
 *
 */
static void inverse_dct(float* dct, float* im, int w, int h)
{
    fftwf_plan r2r;
	r2r = fftwf_plan_r2r_2d(h, w, dct, im, FFTW_REDFT01, FFTW_REDFT01, FFTW_ESTIMATE);
    fftwf_execute(r2r);
    fftwf_destroy_plan(r2r);
    fftwf_cleanup();
}


/**
 *
 *  Weighting the DCT coefficients of an image
 *   using the Fourier transform of the Gaussian function.
 *
 */
static void gaussian_on_dct(float* dctin, float* dctout, int w, int h, float sigma)
{
    float s = sigma*sigma*M_PI*M_PI/2.0;
    for(int m=0; m<h; m++){
        float dm = (float)m/(float)h;
        for(int n = 0; n < w; n++){
            float dn = (float)n/(float)w;
            dctout[m*w+n] = dctin[m*w+n]*exp(-s*(dm*dm+dn*dn));
        }
    }
}


/**
 *
 * Continuous Gaussian convolution of the DCT interpolate.
 *
 */
void gaussconv_dct(float* in, float* out, int w, int h, float sigma)
{
    float* dctin = xmalloc(w*h*sizeof(float));
    float* dctout = xmalloc(w*h*sizeof(float));
    direct_dct(in, dctin, w, h);
    gaussian_on_dct(dctin, dctout, w, h, sigma);
    inverse_dct(dctout, out, w, h);
    xfree(dctin);
    xfree(dctout);
}








/**
 *
 * Compute the Laplacian of an image.
 *
 *  @param gamma: defines the finite difference scheme uses.
 *
 *  The image is extended by symmetrization (to -1/2 and M-1/2)
 *
 */
static void compute_laplacian_scheme(float* in, float* out, int w, int h, float gamma)
{
    assert((0 <= gamma)&&( gamma <= 1));
    int b = 50; // the width of the border

    // extend the image by symmetrization
    int he = h+2*b;
    int we = w+2*b;
    float* padded_in = xmalloc(he*we*sizeof(float));
	for(int i=0;i<he;i++){
		int im = i-b;
		while (im<0) im += 2*h;
		while (im>2*h-1) im -= 2*h;
		if (im>h-1) im = 2*h -1 -im;
		for(int j=0;j<we;j++){
			int jm = j-b;
			while (jm<0){ jm += 2*w;}
			while (jm>2*w-1){ jm -= 2*w;}
			if (jm>w-1) jm = 2*w -1 -jm;
			padded_in[i*we+j] = in[im*w+jm];
		}
	}
    // Laplacian
    for(int i=b; i<h+b; i++){
        for(int j=b; j<w+b; j++){
            // \Delta^{+}
            float lapl_p =   padded_in[(i+1)*we+j] + padded_in[(i-1)*we+j]
                          +   padded_in[i*we+j+1]   + padded_in[i*we+j-1]
                          - 4*padded_in[i*we+j];
            // \Delta^{\times}
            float lapl_t = 0.5*(   padded_in[(i+1)*we+(j+1)] + padded_in[(i-1)*we+(j+1)]
                                  + padded_in[(i+1)*we+(j-1)] + padded_in[(i-1)*we+(j-1)]
                                 -4*padded_in[i*we+j]);
            out[(i-b)*w+(j-b)] = gamma*lapl_p + (1-gamma)*lapl_t;
        }
    }
    xfree(padded_in);
}


/**
 *
 *
 * Lindeberg discrete scale-space smoothing operator.
 *
 *  applies discretized version of the heat equation using the Euler method
 *
 *  @param gamma: defines the finite difference scheme uses.
 *
 */
void gaussconv_lindeberg(float* in, float* out, int w, int h, float sigma, float gamma)
{
    float* lapl = xmalloc(w*h*sizeof(float));

    // Euler iteration number
    int P = 4*(int)( (8*(1-gamma/2)*sigma*sigma) +1);
    // Euler step size
    float dt = sigma*sigma/2/P;

    // initialization
    for(int i=0; i<w*h; i++){
        out[i] = in[i];
    }

    for(int p =0; p<P; p++){
        compute_laplacian_scheme(out, lapl, w, h, gamma);

        // Euler iteration
        for(int i=0; i<w*h; i++){
            out[i] = out[i] + dt* lapl[i];
        }
    }
    xfree(lapl);

}
